#include "woody.h"

int wdy_get_file(char *path, t_map *file) {
	int fd = open(path, O_RDONLY);
	if (fd == -1) {
		return wdy_perror(path);
	}

	file->size = lseek(fd, 0, SEEK_END);
	if (file->size == -1) {
		close(fd);
		return wdy_perror(path);
	}

	file->data = mmap(
			NULL,
			file->size,
			PROT_READ | PROT_WRITE,
			MAP_PRIVATE,
			fd,
			0);
	close(fd);
	if (file->data == MAP_FAILED) {
		return wdy_perror(path);
	}

	return RET_OK;
}

int pack_elf32(t_map file) {
	(void)file;
	return wdy_error("ELF needs to be in 64 bits format");
}

int get_load_segment64(
		Elf64_Ehdr elf_header,
		Elf64_Phdr *program_headers,
		Elf64_Phdr **load_segment) {
	for (int i = 0; i < elf_header.e_phnum; i++) {
		Elf64_Phdr *p_hdr = &program_headers[i];
		if (p_hdr->p_type == PT_LOAD
				&& p_hdr->p_flags & PF_X
				&& p_hdr->p_vaddr <= elf_header.e_entry
				&& p_hdr->p_vaddr + p_hdr->p_filesz > elf_header.e_entry) {
			*load_segment = p_hdr;
			return RET_OK;
		}
	}
	return RET_ERR;
}

// TODO refacto
t_map get_code_cave(t_map file, Elf64_Phdr load_segment) {
	size_t page_size = load_segment.p_align;
	size_t len = load_segment.p_filesz + page_size - load_segment.p_filesz % page_size; 
	size_t offset = load_segment.p_offset;
	unsigned char *seg = fetch(file, offset, len);
	if (!seg) {
		printf("fallback to weird code cave finder\n");
		seg = fetch(file, offset, file.size - offset);
		len = file.size - offset;
		if (!seg) {
			ft_printf("unreachable !!!\n");
		}
	}

	size_t longest = 0;
	size_t longest_i = 0;
	for (size_t i = 0; i < len; i++) {
		if (seg[i] == 0) {
			size_t j = 0;
			while (i + j < len && seg[i + j] == 0) {
				j++;
			}

			if (j > longest) {
				longest_i = i;
				longest = j;
			}
			i += j;

		}
	}

	t_map code_cave;
	code_cave.data = fetch(file, longest_i + offset, longest);
	code_cave.size = longest;
	if (!code_cave.data) {
		printf("unreachable !!!\n");
	}
	return code_cave;
}

int pack_elf64(t_map file) {
	Elf64_Ehdr *elf_header = (Elf64_Ehdr *)fetch(file, 0, sizeof(Elf64_Ehdr));
	if (!elf_header) {
		return wdy_error("cannot fetch elf header");
	}

	size_t phdrs_len = elf_header->e_phnum * elf_header->e_phentsize;
	Elf64_Phdr *program_headers = fetch(file, elf_header->e_phoff, phdrs_len);
	if (!program_headers) {
		return wdy_error("cannot fetch program headers table");
	}

	Elf64_Phdr *load_segment;
	if (get_load_segment64(*elf_header, program_headers, &load_segment) == RET_ERR) {
		return wdy_error("cannot get load segment");
	}

	// TODO the whole section below till the end of the function needs 
	// cleaning and refactoring, better naming also.

	// This should go in a separate function regarding encryption if encryption has been requested.
	/*
	   size_t encryption_block_size = 31; // 32 bits and padding...
	   t_map to_encrypt;

	   to_encrypt.data = fetch(file, load_segment.p_offset, 0);
	// size is a multiple of 31...
	to_encrypt.size = load_segment.p_filesz + encryption_block_size - load_segment.p_filesz % encryption_block_size;
	 */

	t_map code_cave = get_code_cave(file, *load_segment);

	size_t payload_len = 55;
	unsigned char *payload = malloc(payload_len * sizeof(unsigned char));
	ft_memcpy(payload, "\x55\x54\x50\x52\x56\x57\xbf\x01\x00\x00\x00\x48\x8d\x35\x17\x00\x00\x00\xba\x0e\x00\x00\x00\xb8\x01\x00\x00\x00\x0f\x05\x5f\x5e\x5a\x58\x5c\x5d\xe9\xb1\xda\xda\xda\x2e\x2e\x2e\x2e\x57\x4f\x4f\x44\x59\x2e\x2e\x2e\x2e\x0a", payload_len);
	// todo measure this, probably wc/4 in shell script output + hardcoded
	// 55 bytes ^^^ 	

	// same, probably will hardcode it in functions like "get_debug_payload_info"
	// or "get_rsa_payload_info" etc etc...
	size_t jump_i = 36;

	// This should fallback to compression algorithm, or smaller payload (eg rsa->xor)
	if (payload_len > (size_t)code_cave.size) {
		printf("code cave size: %ld (0x%lx) bytes\n", code_cave.size, code_cave.size);
		return wdy_error("payload length exceed code cave size");
	}

	// e_entry because relative to this (where we gonna go)
	// code cave start because thats the start of the code
	// jump_i because thats the index of jump from code cave start
	// 5 because jump_i is 5 bytes long operation (1 byte opcode + 4 byte operand)
	int jump_value = elf_header->e_entry - (code_cave.data - file.data) - jump_i - 5;

	ft_memcpy(payload + jump_i + 1, &jump_value, 4);

	elf_header->e_entry = code_cave.data - file.data;
	//load_segment->p_filesz += payload_len;
	load_segment->p_memsz += payload_len;
	load_segment->p_flags |= PF_W | PF_R;
	ft_memcpy(code_cave.data, payload, payload_len);

	int fd = open("woody", O_WRONLY | O_CREAT, 0755);
	if (fd == -1) {
		return wdy_perror("woody");
	}
	write(fd, file.data, file.size);

	return RET_OK;
}

int pack_elf(t_map file) {	
	// at this point arch is known to be either ELFCLASS32 or ELFCLASS64
	unsigned char arch = ((unsigned char *)file.data)[4];
	if (arch == ELFCLASS64) {
		return pack_elf64(file);
	}
	return pack_elf32(file);
}

int woody_woodpacker(char *path) {
	t_map file;

	if (wdy_get_file(path, &file) == RET_ERR) {
		return RET_ERR;
	}
	if (check_ident(file) == RET_ERR) {
		return RET_ERR;
	}
	return pack_elf(file);
}
